"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Cacheable = exports.localStorageKeys = exports.localStorageMethods = exports.globalKeys = exports.globalMethods = exports.localStorageSet = exports.globalSet = exports.localStorageGet = exports.globalGet = exports.localStorageDelete = exports.globalDelete = exports.localStorageClear = exports.globalClear = exports.localStorageInit = exports.globalInit = void 0;
const CacheableOptions_1 = require("./CacheableOptions");
const CacheableKey_1 = require("./CacheableKey");
const CacheRegistryProvider_1 = require("./registry/CacheRegistryProvider");
const Errors_1 = require("./Errors");
const Symbols_1 = require("./Symbols");
function globalInit(target) {
    CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').initCache(target);
}
exports.globalInit = globalInit;
function localStorageInit(getStore) {
    CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').initialiseStore(getStore);
}
exports.localStorageInit = localStorageInit;
function globalClear(target, methodName) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInit(target, methodName);
    map.clear();
}
exports.globalClear = globalClear;
function localStorageClear(target, methodName) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInit(target, methodName);
    map.clear();
}
exports.localStorageClear = localStorageClear;
function globalDelete(target, methodName, args) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    map.delete(key);
}
exports.globalDelete = globalDelete;
function localStorageDelete(target, methodName, args) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    map.delete(key);
}
exports.localStorageDelete = localStorageDelete;
function globalGet(target, methodName, args) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    return map.get(key);
}
exports.globalGet = globalGet;
function localStorageGet(target, methodName, args) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    return map.get(key);
}
exports.localStorageGet = localStorageGet;
function globalSet(target, methodName, args, val) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    map.set(key, val);
}
exports.globalSet = globalSet;
function localStorageSet(target, methodName, args, val) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInit(target, methodName);
    const key = buildCacheKey(args, `${target.constructor.name}::${methodName}`);
    map.set(key, val);
}
exports.localStorageSet = localStorageSet;
function globalMethods(target) {
    return new Set(CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInitCache(target).keys());
}
exports.globalMethods = globalMethods;
function globalKeys(target, methodName) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('GLOBAL').getOrInit(target, methodName);
    return map.keys();
}
exports.globalKeys = globalKeys;
function localStorageMethods(target) {
    return new Set(CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInitCache(target).keys());
}
exports.localStorageMethods = localStorageMethods;
function localStorageKeys(target, methodName) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope('LOCAL_STORAGE').getOrInit(target, methodName);
    return map.keys();
}
exports.localStorageKeys = localStorageKeys;
function Cacheable(options) {
    return (_target, _propertyKey, descriptor) => {
        if (descriptor.value != undefined) {
            descriptor.value = wrap(descriptor.value, (0, CacheableOptions_1.optionsWithDefaults)(options));
        }
        else if (descriptor.get != undefined) {
            descriptor.get = wrap(descriptor.get, (0, CacheableOptions_1.optionsWithDefaults)(options));
        }
        else {
            throw new Errors_1.UncacheablePropertyError('Only put a Cacheable() decorator on a method or get accessor.');
        }
    };
}
exports.Cacheable = Cacheable;
function wrap(originalMethod, options) {
    return function (...args) {
        return cacheOriginalMethod.apply(this, [originalMethod, options, args]);
    };
}
function cacheOriginalMethod(originalMethod, options, args) {
    const map = CacheRegistryProvider_1.CacheRegistryProvider.forScope(options.scope).getOrInit(this, originalMethod.name, options);
    const cacheKey = buildCacheKey(args, `${this.constructor.name}::${originalMethod.name}`);
    if (map.has(cacheKey)) {
        return map.get(cacheKey);
    }
    else {
        const returnValueOrPromise = originalMethod.apply(this, args);
        function returnValueHandler(returnValue) {
            if (returnValue != undefined || options.cacheUndefined === true) {
                map.set(cacheKey, returnValue, options.ttl);
            }
            return returnValue;
        }
        if (isPromiseLike(returnValueOrPromise)) {
            return returnValueOrPromise.then(returnValueHandler);
        }
        else {
            return returnValueHandler(returnValueOrPromise);
        }
    }
}
function buildCacheKey(args, symbolName) {
    if (!args || !args.length) {
        return Symbols_1.NoArgsCacheKey;
    }
    const argCacheKeys = args.map((it, index) => {
        if (it === null) {
            return Symbols_1.NullValueCacheKey;
        }
        else if (it === undefined) {
            return Symbols_1.UndefinedValueCacheKey;
        }
        else if (typeof it === 'object' && (0, CacheableKey_1.implementsCacheableKey)(it)) {
            return it.cacheKey();
        }
        else {
            try {
                return JSON.stringify(it);
            }
            catch (_a) {
                throw new Errors_1.UncacheableArgumentError([
                    'Cannot cache: "' + symbolName + '".',
                    'To serve as a cache key, a parameter must be serializable to JSON,',
                    'and should return a unique value.',
                    'The argument at index ' + index + ' does not.',
                    'Alternatively, consider providing a hash function, by implementing the CacheableKey interface.',
                ].join(' '));
            }
        }
    });
    if (argCacheKeys.length === 1) {
        return argCacheKeys[0];
    }
    else {
        return argCacheKeys.reduce((previousValue, currentValue) => `${String(previousValue)}_${String(currentValue)}`);
    }
}
function isPromiseLike(value) {
    return ((typeof value === 'function' || typeof value === 'object') &&
        value !== null &&
        typeof value.then === 'function');
}
//# sourceMappingURL=Cacheable.js.map